use crate::command::Command;

use std::thread;
use std::error::Error;
use std::time::{Duration, Instant};

use embedded_hal::serial::{Read, Write};
use embedded_hal::digital::v2::OutputPin;
use esp_idf_hal::serial::{self, Rx, Tx};

pub type Result<T> = std::result::Result<T, ModemError>;

pub struct Modem<UART: serial::Uart> {
    rx: IterableRx<UART>,
    tx: Tx<UART>,
}

#[derive(Debug)]
pub enum ModemError {
    CommandError(String),
    SetupError(String),
    ReadError,
    TimeoutError,
}

impl Error for ModemError {}

impl std::fmt::Display for ModemError {
    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
        write!(f, "{:?}", self)
    }
}

pub struct IterableRx<UART: serial::Uart> {
    inner: Rx<UART>,
    timeout: Option<Duration>,
}

impl<UART: serial::Uart> IterableRx<UART> {
    fn reset(&mut self, timeout: Duration) -> &mut Self {
        self.timeout = Some(timeout);
        self
    }
}

impl<UART: serial::Uart> Iterator for IterableRx<UART> {
    type Item = u8;

    /// `nb` returns Ok(byte), or one of Err(WouldBlock) and Err(Other) which isn't of anyone's
    /// interest, so the retry mechanism is triggered on _any_ error every 200ms until a byte is
    /// received, or the timeout is reached.
    fn next(&mut self) -> Option<Self::Item> {
        let now = Instant::now();
        loop {
            let timeout = self.timeout.unwrap_or(Duration::from_millis(0));
            match self.inner.read() {
                Ok(b) => {
                    break Some(b)
                },
                _ => {
                    if now.elapsed() > timeout {
                        break None
                    }
                    thread::sleep(Duration::from_millis(200));
                    self.timeout = Some(timeout.saturating_sub(now.elapsed()));
                }
            }
        }
    }
}

impl<UART: serial::Uart> Modem<UART> {
    pub fn new(tx: Tx<UART>, rx: Rx<UART>) -> Self {
        Self {
            rx: IterableRx { inner: rx, timeout: None },
            tx,
        }
    }

    /// Initialize the modem (sim800l in this case). The initialization process sets all pins in the
    /// required state so that the modem is turned on, then resets it a couple of times (beats me) and
    /// sleeps for 3 seconds, which is enough for the modem to come online.
    ///
    /// Below is an example for sim800l pins on a LilyGo TTGO T-Call.
    ///
    /// # Examples
    ///
    /// ```
    /// let modem_pwrkey = dp.pins.gpio4.into_output().unwrap();
    /// let modem_rst = dp.pins.gpio5.into_output().unwrap();
    /// let modem_power = dp.pins.gpio23.into_output().unwrap();
    ///
    /// modem::init(modem_pwrkey, modem_rst, modem_power);
    /// ```
    pub fn init(&mut self, mut pwrkey: impl OutputPin, mut rst: impl OutputPin, mut power: impl OutputPin) -> Result<()> {
        println!("Turning SIM800L on ...");
        power.set_high().map_err(|_| ModemError::SetupError("Error setting POWER to high.".to_string()))?;
        rst.set_high().map_err(|_| ModemError::SetupError("Error setting RST to high.".to_string()))?;
        // Pull down PWRKEY for more than 1 second according to manual requirements
        pwrkey.set_high().map_err(|_| ModemError::SetupError("Error setting PWRKEY to high.".to_string()))?;
        thread::sleep(Duration::from_millis(100));
        pwrkey.set_low().map_err(|_| ModemError::SetupError("Error setting PWRKEY to low.".to_string()))?;
        thread::sleep(Duration::from_millis(1000));
        pwrkey.set_high().map_err(|_| ModemError::SetupError("Error setting PWRKEY to high.".to_string()))?;
        println!("Waiting 3s for sim module to come online ...");
        thread::sleep(Duration::from_millis(3000));
        Ok(())
    }

    /// Reads a whole line (that ends with \\n) within the given `timeout` passed on input.
    ///
    fn read_line(&mut self, timeout: Duration) -> Result<String> {
        let line: String = self.rx.reset(timeout)
            .map(|b| char::from(b))
            .take_while(|c| *c != '\n')
            .collect();

        // A necessary check because the actual timeout is in the Iterator implementation. The
        // iterator exits when the returned Item is None, which happens when there's no data in
        // the serial port and the timeout is breached.
        //
        // This check here is tested on sim800l and works only because the modem has \r\n as CRLF.
        if line.ends_with("\r") {
            Ok(line)
        }
        else {
            Err(ModemError::ReadError)
        }
    }

    /// Reads the serial RX until a \\n char is encoutered, or a timeout is reached. The timeout is
    /// provided on input via the `timeout` argument. The first argument `contains` is checked
    /// against a line in the response, if it's there the reading stops.
    ///
    /// If `contains` is `None`, the first line only is returned in the response. If it's
    /// `Some(match_txt)`, then the end of the response is matched against `match_txt`.
    fn read_response(&mut self, contains: Option<String>, timeout: Duration) -> Result<String> {
        let mut response = String::new();
        let start = Instant::now();
        let match_text: String = contains.unwrap_or("\n".to_string());

        loop {
            let rdln = self.read_line(timeout.saturating_sub(start.elapsed()));
            if let Ok(line) = rdln {
                println!("Read {} bytes from serial ({})", line.len(), line);
                response.push_str(&line);
                if line.contains("ERROR") || line.contains(&match_text) {
                    println!("Found match {} for line {} ; exiting response reader now ...", match_text, line);
                    println!("-----------------------------------------------------------");
                    break Ok(response.to_string())
                }
            } else {
                println!("-----------------------------------------------------------");
                println!("Read line {:?}", rdln);
                break Err(ModemError::TimeoutError)
            }
        }
    }

    fn send(&mut self, b: u8) -> Result<()> {
        nb::block!(self.tx.write(b))
            .map_err(|_| ModemError::CommandError(format!("error writing {} to serial", b)))?;
        Ok(())
    }

    fn send_bytes(&mut self, payload: &[u8]) -> Result<()> {
        for b in payload.iter() {
            self.send(*b)?;
        }
        self.send('\r' as u8)?;
        Ok(())
    }

    fn send_command(&mut self, cmd: Command) -> Result<String> {
        println!("-----------------------------------------------------------");
        println!("Sending {} ...", cmd.text);
        let _ = self.send_bytes(cmd.text.as_bytes())?;
        self.read_response(cmd.contains, cmd.timeout)
    }

    fn send_command_data(&mut self, cmd: Command, payload: &str) -> Result<String> {
        println!("-----------------------------------------------------------");
        println!("Sending {} ...", cmd.text);
        let _ = self.send_bytes(cmd.text.as_bytes())?;

        let prompt: String = self.rx.reset(cmd.timeout)
            .map(|b| char::from(b))
            .take_while(|c| *c != ' ')
            .collect();

        if prompt != ">".to_string() {
            println!("invalid prompt: {}", prompt);
        }
        let _ = self.send_bytes(payload.as_bytes())?;
        self.read_response(cmd.contains, cmd.timeout)
    }

    pub fn get_ip_addr(&mut self) -> Result<String> {
        self.send_command(Command::getbear())
    }

    pub fn get_local_ip_addr(&mut self) -> Result<()> {
        let _ = self.send_command(Command::get_local_ip_addr())?;
        Ok(())
    }

    pub fn connect_to_gprs_ap(&mut self, apn: &str, username: &str, password: &str)-> Result<()> {
        println!("init gprs ...");
        let _ = self.send_command(Command::gprs_init())?;

        println!("setting up gprs credentials for apn {}, {}:{})", apn, username, password);

        let _ = self.send_command(Command::gprs_set_apn(apn))?;
        let _ = self.send_command(Command::gprs_set_user(username))?;
        let _ = self.send_command(Command::gprs_set_pwd(password))?;

        println!("open gprs ...");
        let _ = self.send_command(Command::gprs_open())?;

        Ok(())
    }

    pub fn info(&mut self)-> Result<String> {
        self.send_command(Command::modem_info())
    }

    pub fn probe(&mut self)-> Result<String> {
        self.send_command(Command::probe())
    }

    pub fn is_gprs_attached(&mut self)-> Result<bool> {
        let res = self.send_command(Command::is_gprs_attached())?;
        Ok(res.contains("+CGATT: 1"))
    }

    pub fn ping(&mut self, domain: &str)-> Result<()> {
        self.send_command(Command::ping(domain))?;
        Ok(())
    }

    pub fn tcp_is_ssl_enabled(&mut self) -> Result<bool> {
        let res = self.send_command(Command::tcp_ssl_check())?;
        Ok(res.contains("+CIPSSL: (1)"))
    }

    pub fn tcp_ssl_disable(&mut self) -> Result<()> {
        let _ = self.send_command(Command::tcp_ssl_disable())?;
        Ok(())
    }

    pub fn tcp_connect(&mut self, addr: &str, port: u16) -> Result<()> {
        self.send_command(Command::tcp_connect(addr, port))?;
        Ok(())
    }

    pub fn tcp_set_quick_mode(&mut self, mode: bool) -> Result<()> {
        self.send_command(Command::tcp_set_quick_mode(mode))?;
        Ok(())
    }

    pub fn tcp_send(&mut self, payload: &str) -> Result<()> {
        self.send_command(Command::tcp_send_size(payload.len()))?;
        self.send_command_data(Command::tcp_send(), payload)?;
        Ok(())
    }
}
